import pathlib
import timeit

import numpy as np
from karabo.bound import (
    BOOL_ELEMENT,
    NODE_ELEMENT,
    FLOAT_ELEMENT,
    KARABO_CLASSINFO,
    STRING_ELEMENT,
    UINT32_ELEMENT,
    Hash,
    Schema,
)
from karabo.common.states import State

from . import utils
from ._version import version as deviceVersion
from .agipd_gpu import AgipdGainMode, AgipdGpuRunner, BadPixelValues, CorrectionFlags
from .base_correction import BaseCorrection
from .calcat_utils import AgipdCalcatFriend, AgipdConstants


@KARABO_CLASSINFO("AgipdCorrection", deviceVersion)
class AgipdCorrection(BaseCorrection):
    # subclass *must* set these attributes
    _correction_flag_class = CorrectionFlags
    _correction_slot_names = (
        ("thresholding", CorrectionFlags.THRESHOLD),
        ("offset", CorrectionFlags.OFFSET),
        ("relGainPc", CorrectionFlags.REL_GAIN_PC),
        ("relGainXray", CorrectionFlags.REL_GAIN_XRAY),
        ("badPixels", CorrectionFlags.BPMASK),
    )
    _gpu_runner_class = AgipdGpuRunner
    _schema_cache_slots = BaseCorrection._schema_cache_slots | {"sendGainMap"}

    @staticmethod
    def expectedParameters(expected):
        super(AgipdCorrection, AgipdCorrection).expectedParameters(expected)
        AgipdCalcatFriend.add_schema(expected)
        # TODO: encapsulate correction configuration subschema
        for slot_name, _ in AgipdCorrection._correction_slot_names:
            (
                BOOL_ELEMENT(expected)
                .key(f"corrections.available.{slot_name}")
                .readOnly()
                .initialValue(False)
                .commit(),
                BOOL_ELEMENT(expected)
                .key(f"corrections.enabled.{slot_name}")
                .assignmentOptional()
                .defaultValue(False)
                .reconfigurable()
                .commit(),
                BOOL_ELEMENT(expected)
                .key(f"corrections.preview.{slot_name}")
                .assignmentOptional()
                .defaultValue(False)
                .reconfigurable()
                .commit(),
            )

        (
            BOOL_ELEMENT(expected)
            .key("corrections.overrideMdAdditionalOffset")
            .displayedName("Override md_additional_offset")
            .assignmentOptional()
            .defaultValue(False)
            .reconfigurable()
            .commit(),
            FLOAT_ELEMENT(expected)
            .key("corrections.mdAdditionalOffset")
            .displayedName("Value for md_additional_offset (if overriding)")
            .description(
                "Normally, md_additional_offset (part of relative gain correction) is "
                "computed when loading SlopesPC. In case you want to use a different "
                "value (global for all medium gain pixels), you can specify it here "
                "and set corrections.overrideMdAdditionalOffset to True."
            )
            .assignmentOptional()
            .defaultValue(0)
            .reconfigurable()
            .commit(),
        )
        (
            STRING_ELEMENT(expected)
            .key("gainMode")
            .displayedName("Gain mode")
            .assignmentOptional()
            .defaultValue("ADAPTIVE_GAIN")
            .options("ADAPTIVE_GAIN,FIXED_HIGH_GAIN,FIXED_MEDIUM_GAIN,FIXED_LOW_GAIN")
            .commit(),
            BOOL_ELEMENT(expected)
            .key("sendGainMap")
            .displayedName("Send gain map on dataOutput")
            .assignmentOptional()
            .defaultValue(False)
            .reconfigurable()
            .commit(),
        )
        # TODO: hook this up to actual correction done
        # NOTE: wanted as table, could not make readonly rows of reconfigurable bools
        (
            NODE_ELEMENT(expected)
            .key("corrections.badPixelFlagsToUse")
            .displayedName("Bad pixel flags to use")
            .description(
                "The booleans under this node allow for selecting a subset of bad "
                "pixel values to take into account when doing bad pixel masking. "
                "Upon updating these flags, the map used for bad pixel masking will "
                "be ANDed with this selection. TEMPORARY NOTE: if you want to toggle "
                "a disabled flag back on, please reload constants for this to take "
                "effect (will be triggered automatically in future version)."
            )
            .commit(),
        )
        for field in BadPixelValues:
            (
                BOOL_ELEMENT(expected)
                .key(f"corrections.badPixelFlagsToUse.{field.name}")
                .assignmentOptional()
                .defaultValue(True)
                .reconfigurable()
                .commit()
            )
        (
            STRING_ELEMENT(expected)
            .key("corrections.badPixelMaskValue")
            .displayedName("Bad pixel masking value")
            .description(
                "Any pixels masked by the bad pixel mask will have their value "
                "replaced with this. Note that this setting is evaluated as a string; "
                "use float('nan') to get NaN value."
            )
            .assignmentOptional()
            .defaultValue("float('nan')")
            .reconfigurable()
            .commit()
        )

    @property
    def input_data_shape(self):
        return (
            self.get("dataFormat.memoryCells"),
            2,
            self.get("dataFormat.pixelsX"),
            self.get("dataFormat.pixelsY"),
        )

    @property
    def output_data_shape(self):
        return utils.shape_after_transpose(
            (
                self.get("dataFormat.memoryCells"),
                self.get("dataFormat.pixelsX"),
                self.get("dataFormat.pixelsY"),
            ),
            self._output_transpose,
        )

    def __init__(self, config):
        super().__init__(config)
        # TODO: consider putting this initialization in base class
        self.calibration_constant_manager = AgipdCalcatFriend(
            self, pathlib.Path.cwd() / "calibration-client-secrets.json"
        )
        # TODO: different gpu runner for fixed gain mode
        self.gain_mode = AgipdGainMode[config.get("gainMode")]
        self.bad_pixel_mask_value = eval(config.get("corrections.badPixelMaskValue"))
        self._gpu_runner_init_args = {
            "gain_mode": self.gain_mode,
            "bad_pixel_mask_value": self.bad_pixel_mask_value,
        }

        self._output_transpose = {
            "pixels-fast": None,
            "memorycells-fast": (2, 1, 0),
            "no-reshape": None,
        }[config.get("dataFormat.outputAxisOrder")]
        self._update_shapes()

        # configurability: overriding md_additional_offset
        if config.get("corrections.overrideMdAdditionalOffset"):
            self._override_md_additional_offset = config.get(
                "corrections.mdAdditionalOffset"
            )
        else:
            self._override_md_additional_offset = None

        # configurability: disabling subset of bad pixel masking bits
        self._has_updated_bad_pixel_selection = False
        self._update_bad_pixel_selection()

        self.updateState(State.ON)
        self.KARABO_SLOT(self.loadMostRecentConstants)

    def process_input(self, data, metadata):
        """Registered for dataInput, handles all processing and sending"""

        if not self._schema_cache["doAnything"]:
            if self._schema_cache["state"] is State.PROCESSING:
                self.updateState(State.ACTIVE)
            return

        source = metadata.get("source")

        if source not in self.sources:
            self.log.INFO(f"Ignoring unknown source {source}")
            return

        # TODO: what are these empty things for?
        if not data.has("image"):
            # self.log.INFO("Ignoring hash without image node")
            return

        time_start = timeit.default_timer()

        train_id = metadata.getAttribute("timestamp", "tid")
        cell_table = np.squeeze(data.get("image.cellId"))
        assert isinstance(cell_table, np.ndarray), "image.cellId should be ndarray"
        if len(cell_table.shape) == 0:
            msg = "cellId had 0 dimensions. DAQ may not be sending data."
            self.set("status", msg)
            self.log.WARN(msg)
            return
        # original shape: memory_cell, data/raw_gain, x, y
        image_data = data.get("image.data")
        if image_data.shape[0] != self._schema_cache["dataFormat.memoryCells"]:
            self.set(
                "status", f"Updating input shapes based on received {image_data.shape}"
            )
            self.set("dataFormat.memoryCells", image_data.shape[0])
            with self._buffer_lock:
                # TODO: pulse filter update after reimplementation
                self._update_shapes()

        if not self._schema_cache["state"] is State.PROCESSING:
            self.updateState(State.PROCESSING)
            self.set("status", "Processing data")
        if self._state_reset_timer is None:
            self._state_reset_timer = utils.DelayableTimer(
                timeout=self._schema_cache["processingStateTimeout"],
                callback=self._reset_state_from_processing,
            )
        else:
            self._state_reset_timer.set_timeout(
                self._schema_cache["processingStateTimeout"]
            )

        correction_cell_num = self._schema_cache["constantParameters.memoryCells"]
        do_generate_preview = (
            train_id % self._schema_cache["preview.trainIdModulo"] == 0
            and self._schema_cache["preview.enable"]
        )

        with self._buffer_lock:
            # cell_table = cell_table[self.pulse_filter]
            pulse_table = np.squeeze(data.get("image.pulseId"))  # [self.pulse_filter]
            cell_table_max = np.max(cell_table)
            # TODO: all this checking and warning can go into GPU runner
            if cell_table_max >= correction_cell_num:
                msg = (
                    f"Max cell ID ({cell_table_max}) exceeds range for loaded "
                    f"constants ({correction_cell_num} cells). Some frames will not be "
                    "corrected."
                )
                self.log.WARN(msg)
                self.set("status", msg)

            self.gpu_runner.load_data(image_data)
            buffer_handle, buffer_array = self._shmem_buffer.next_slot()
            self.gpu_runner.load_cell_table(cell_table)
            self.gpu_runner.correct(self._correction_flag_enabled)
            self.gpu_runner.reshape(out=buffer_array)
            # after reshape, data for dataOutput is now safe in its own buffer
            if do_generate_preview:
                if self._correction_flag_enabled != self._correction_flag_preview:
                    self.gpu_runner.correct(self._correction_flag_preview)
                preview_slice_index = self._schema_cache["preview.pulse"]
                if preview_slice_index >= 0:
                    # look at pulse_table to find which index this pulse ID is in
                    # TODO: move this to GPU
                    pulse_id_found = np.where(pulse_table == preview_slice_index)[0]
                    if len(pulse_id_found) == 0:
                        pulse_found_instead = pulse_table[0]
                        msg = (
                            f"Pulse {preview_slice_index} not found in "
                            f"image.pulseId, arbitrary pulse "
                            f"{pulse_found_instead} will be shown."
                        )
                        self.log.WARN(msg)
                        self.set("status", msg)
                        preview_slice_index = 0
                    else:
                        preview_slice_index = pulse_id_found[0]
                preview_raw, preview_corrected = self.gpu_runner.compute_preview(
                    preview_slice_index
                )

        data.set("image.data", buffer_handle)
        data.set("image.cellId", cell_table[:, np.newaxis])
        data.set("image.pulseId", pulse_table[:, np.newaxis])
        data.set("calngShmemPaths", ["image.data"])
        self._write_output(data, metadata)
        if do_generate_preview:
            self._write_combiner_preview(
                preview_raw, preview_corrected, train_id, source
            )

        # update rate etc.
        self._buffered_status_update.set("trainId", train_id)
        # self._rate_tracker.update()
        time_spent = timeit.default_timer() - time_start
        self._buffered_status_update.set(
            "performance.lastProcessingDuration", time_spent * 1000
        )

    def loadMostRecentConstants(self):
        self.flush_constants()
        self.calibration_constant_manager.flush_constants()
        for constant in AgipdConstants:
            self.calibration_constant_manager.get_constant_version_and_call_me_back(
                constant, self._load_constant_to_gpu
            )

    def _load_constant_to_gpu(self, constant, constant_data):
        # TODO: encode correction / constant dependencies in a clever way
        if constant is AgipdConstants.ThresholdsDark:
            if self.gain_mode is not AgipdGainMode.ADAPTIVE_GAIN:
                self.log.INFO("Loaded ThresholdsDark ignored due to fixed gain mode")
                return
            self.gpu_runner.load_thresholds(constant_data)
            if not self.get("corrections.available.thresholding"):
                self.set("corrections.available.thresholding", True)
                self.set("corrections.enabled.thresholding", True)
                self.set("corrections.preview.thresholding", True)
        elif constant is AgipdConstants.Offset:
            self.gpu_runner.load_offset_map(constant_data)
            if not self.get("corrections.available.offset"):
                self.set("corrections.available.offset", True)
                self.set("corrections.enabled.offset", True)
                self.set("corrections.preview.offset", True)
        elif constant is AgipdConstants.SlopesPC:
            self.gpu_runner.load_rel_gain_pc_map(constant_data)
            if not self.get("corrections.available.relGainPc"):
                self.set("corrections.available.relGainPc", True)
                self.set("corrections.enabled.relGainPc", True)
                self.set("corrections.preview.relGainPc", True)
            if self._override_md_additional_offset is not None:
                self.gpu_runner.md_additional_offset_gpu.fill(
                    self._override_md_additional_offset
                )
        elif constant is AgipdConstants.SlopesFF:
            self.gpu_runner.load_rel_gain_ff_map(constant_data)
            if not self.get("corrections.available.relGainXray"):
                self.set("corrections.available.relGainXray", True)
                self.set("corrections.enabled.relGainXray", True)
                self.set("corrections.preview.relGainXray", True)
        elif "BadPixels" in constant.name:
            self.gpu_runner.load_bad_pixels_map(
                constant_data, override_flags_to_use=self._override_bad_pixel_flags
            )
            if not self.get("corrections.available.badPixels"):
                self.set("corrections.available.badPixels", True)
                self.set("corrections.enabled.badPixels", True)
                self.set("corrections.preview.badPixels", True)

        self._update_correction_flags()

    def _update_pulse_filter(self, filter_string):
        """Called whenever the pulse filter changes, typically followed by
        _update_shapes"""

        if filter_string.strip() == "":
            new_filter = np.arange(self.get("dataFormat.memoryCells"), dtype=np.uint16)
        else:
            new_filter = np.array(eval(filter_string), dtype=np.uint16)
        assert np.max(new_filter) < self.get("dataFormat.memoryCells")
        self.pulse_filter = new_filter

    def _update_bad_pixel_selection(self):
        selection = 0
        for field in BadPixelValues:
            if self.get(f"corrections.badPixelFlagsToUse.{field.name}"):
                selection |= field
        self._override_bad_pixel_flags = selection

    def preReconfigure(self, config):
        super().preReconfigure(config)
        if any(
            path.startswith("corrections.badPixelFlagsToUse")
            for path in config.getPaths()
        ):
            self._has_updated_bad_pixel_selection = False
        if config.has("corrections.badPixelMaskValue"):
            self.bad_pixel_mask_value = eval(
                config.get("corrections.badPixelMaskValue")
            )
            self.gpu_runner.set_bad_pixel_mask_value(self.bad_pixel_mask_value)

    def postReconfigure(self):
        super().postReconfigure()

        if self.get("corrections.overrideMdAdditionalOffset"):
            self._override_md_additional_offset = self.get(
                "corrections.mdAdditionalOffset"
            )
            self.gpu_runner.md_additional_offset_gpu.override_md_additional_offset(
                self._override_md_additional_offset
            )
        else:
            self._override_md_additional_offset = None

        if not self._has_updated_bad_pixel_selection:
            self._update_bad_pixel_selection()
            self.gpu_runner.override_bad_pixel_flags_to_use(
                self._override_bad_pixel_flags
            )
